
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '35.6.0',
    'branch_nick': 'release-35-LTS',
    'revision_id': '76a17347408868f403f602b5bf626c9744620829',
    'revno': 13190
}
